"""
Test settings API
"""

import unittest
from egnyte import configuration, client

CONFIG_NAME = 'test_config.json'

class TestSettings(unittest.TestCase):
    settings = None

    def setUp(self):
        self.config = configuration.load(CONFIG_NAME)
        self.egnyte = client.EgnyteClient(self.config)

    def test_audit_settings(self):
        audit = self.__fetch_settings()["audit"]

        self.assertIn("archive_audit_reports", audit)

    def test_file_system_settings(self):
        file_system = self.__fetch_settings()["file_system"]

        self.assertIn("locale", file_system)
        self.assertIn("total", file_system)
        self.assertIn("used", file_system)

    def test_users_settings(self):
        users = self.__fetch_settings()["users"]

        self.assertIn("licenses", users)
        self.assertIn("sso_enabled", users)
        self.assertIn("power_user_rights", users)

    def test_links_settings(self):
        links = self.__fetch_settings()["links"]

        self.assertIn("folder_links", links)
        self.assertIn("file_links", links)
        self.assertIn("expiration", links)

    def __fetch_settings(self):
        if TestSettings.settings is None:
            TestSettings.settings = self.egnyte.settings
        return TestSettings.settings



#!/usr/bin/env python
from __future__ import print_function, unicode_literals

import getpass
import argparse
import json
import sys
import datetime
import codecs
from contextlib import closing

from egnyte import client, configuration, exc, base


parser_kwargs = dict(formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50))


def create_main_parser():
    main = argparse.ArgumentParser(prog="python -m egnyte", **parser_kwargs)
    main.add_argument("-c", "--config-path", help="Path to config file")
    main.add_argument('-v', '--verbose', action='count', dest='verbosity', help="Be more verbose. Can be repeated for debugging", default=0)
    main.add_argument('--impersonate', metavar="USERNAME", help="Impersonate another user (username or email)", default=None)

    subparsers = main.add_subparsers()

    parser_config = subparsers.add_parser('config', help='commands related to configuration', **parser_kwargs)
    subparsers_config = parser_config.add_subparsers()

    parser_config_show = subparsers_config.add_parser('show', help="show configuration", **parser_kwargs)
    parser_config_show.set_defaults(command="config_show")

    parser_config_create = subparsers_config.add_parser('create', help='create a new configuration file', **parser_kwargs)
    parser_config_create.set_defaults(command="config_create")

    parser_config_update = subparsers_config.add_parser('update', help='update a configuration file', **parser_kwargs)
    parser_config_update.set_defaults(command="config_update")

    parser_config_token = subparsers_config.add_parser('token', help='generate a new access token and store it in config file', **parser_kwargs)
    parser_config_token.set_defaults(command="config_token")

    parser_token = subparsers.add_parser('token', help='generate a new access token and print it', **parser_kwargs)
    parser_token.set_defaults(command='token')

    parser_test = subparsers.add_parser('test', help='test if config is correct (connects to service)', **parser_kwargs)
    parser_test.set_defaults(command='test')

    for parser, required in [(parser_config_create, True), (parser_config_update, False), (parser_token, False)]:
        parser.add_argument('-d', '--domain', required=required, help='domain name')
        parser.add_argument('-l', '--login', required=False, help='login')
        parser.add_argument('-p', '--password', required=False, help='password')
        parser.add_argument('-k', '--key', dest='api_key', required=required, help='API key')

    for parser in (parser_config_create, parser_config_update):
        parser.add_argument('-t', '--token', dest='access_token', required=False, help='API access token')
        parser.add_argument('-T', '--timeout', required=False, help='Request timeout')

    # Audit generator

    parser_audit = subparsers.add_parser('audit', help='generate audit reports', **parser_kwargs)
    subparsers_audit = parser_audit.add_subparsers()

    parser_audit_files = subparsers_audit.add_parser('files', help="create Files report", **parser_kwargs)
    parser_audit_files.set_defaults(command="audit_files")

    parser_audit_logins = subparsers_audit.add_parser('logins', help="create Logins report", **parser_kwargs)
    parser_audit_logins.set_defaults(command="audit_logins")

    parser_audit_permissions = subparsers_audit.add_parser('permissions', help="create Permissions report", **parser_kwargs)
    parser_audit_permissions.set_defaults(command="audit_permissions")

    parser_audit_get = subparsers_audit.add_parser('get', help="get a previously generated report", **parser_kwargs)
    parser_audit_get.set_defaults(command="audit_get")

    # Common options
    for parser in (parser_audit_files, parser_audit_logins, parser_audit_permissions, parser_audit_get):
        parser.add_argument('--save', required=False, default=None, help="File to save to the report to (default is standard output)")

    for parser in (parser_audit_files, parser_audit_logins, parser_audit_permissions):
        parser.add_argument('--format', required=False, default="csv", help="Report type (json or csv. Default is csv)")
        parser.add_argument('--start', required=False, default='yesterday', help='Start date (YYYY-MM-DD)')
        parser.add_argument('--end', required=False, default='today', help='End date (YYYY-MM-DD)')

    parser_audit_files.add_argument('--folder', required=False, action='append', default=None, help="Absolute folder path for the destination folder. 'folder' or 'file' is required. Can be used multiple times")
    parser_audit_files.add_argument('--file', required=False, default=None, help="Absolute folder path for the destination file, wildcards allowed. 'folder' or 'file' is required")
    parser_audit_files.add_argument('--users', required=False, default=None, help='Users to report on (comma separated list, default is all)')
    parser_audit_files.add_argument('--transaction_type', required=False, default=None, help="""
    Transaction type: upload, download, preview, delete, copy, move, restore_trash, delete_trash, create_link, delete_link, download_link
    (comma separated list, default is all""")

    parser_audit_logins.add_argument('--events', required=True, help="Event types: logins, logouts, account_lockouts, password_resets, failed_attempts (comma separated list)")
    parser_audit_logins.add_argument('--access-points', required=False, default=None, help="Access points to cover: Web, FTP, Mobile (comma separated list, default is all)")
    parser_audit_logins.add_argument('--users', required=False, default=None, help='Users to report on (comma separated list, default is all)')

    parser_audit_permissions.add_argument('--assigners', required=True, help='Permission assigners (comma separated list)')
    parser_audit_permissions.add_argument('--folder', required=True, action='append', default=None, help="Absolute folder path for the destination folder. Can be used multiple times")
    parser_audit_permissions.add_argument('--users', required=False, default=None, help='Users to report on (comma separated list)')
    parser_audit_permissions.add_argument('--groups', required=False, default=None, help='Groups to report on (comma separated list)')

    parser_audit_get.add_argument('--id', type=int, required=True, help="Id of the report")

    parser_upload = subparsers.add_parser('upload', help='send files to Egnyte', **parser_kwargs)
    parser_upload.set_defaults(command="upload")

    parser_upload.add_argument('paths', nargs='+', help="Paths (files to directories) to upload")
    parser_upload.add_argument('target', help="Path in Cloud File System to upload to")
    parser_upload.add_argument('-x', '--exclude', action='append', default=None, help='Exclude items that match this glob pattern')

    parser_download = subparsers.add_parser('download', help='download files from Egnyte', **parser_kwargs)
    parser_download.set_defaults(command="download")

    parser_download.add_argument('paths', nargs='+', help="Paths (files to directories) to download")
    parser_download.add_argument('--target', help="Local directory to put downloaded files and directories in", default='.')
    parser_download.add_argument('--overwrite', action='store_const', const=True, default=False, help="Delete local files and directories that conflict with cloud content")

    parser_settings = subparsers.add_parser('settings', help='show domain settings', **parser_kwargs)
    parser_settings.set_defaults(command="settings")

    parser_search = subparsers.add_parser('search', help='search for files', **parser_kwargs)
    parser_search.set_defaults(command="search")
    parser_search.add_argument('query', help='Search query')
    parser_search.add_argument('--mtime_from', help="Minimim modification date", default=None)
    parser_search.add_argument('--mtime_to', help="Maximum modification date", default=None)
    parser_search.add_argument('--folder', help="Limit search to a specified folder", default=None)

    parser_events = subparsers.add_parser('events', help='show events from the domain', **parser_kwargs)
    parser_events.set_defaults(command="events")
    parser_events.add_argument('--start', type=int, help="Starting event id. Default or 0 - last seen event. Negative numbers are counter backwards from last event", default=None)
    parser_events.add_argument('--stop', type=int, help="Stop event id. Default - poll indefinitely. 0 means last event. Negative numbers are counter backwards from last event", default=None)
    parser_events.add_argument('--type', action='append', help="Limit to events of specific type", default=None)
    parser_events.add_argument('--folder', help="Limit to events in specific folder and it's subfolders", default=None)
    parser_events.add_argument('--suppress', help="Skip events caused by this app or user. Valid values: app, user.", default=None)





    return main

def to_json(obj):
    return {k:v for (k, v) in obj.__dict__.items() if not k.startswith('_')}

class Commands(object):
    _config = None
    config_keys = ('login', 'password', 'domain', 'api_key', 'access_token', 'timeout')
    STATUS_CMD_NOT_FOUND = 1
    STATUS_API_ERROR = 2
    INFO = 1
    DEBUG = 2

    def load_config(self):
        if self._config is None:
            self._config = configuration.load(self.args.config_path)
        return self._config

    def save_config(self):
        return configuration.save(self.config, self.args.config_path)

    config = property(load_config)

    def __init__(self, args):
        self.args = args

    @property
    def info(self):
        """If verbosity is INFO or better"""
        return self.args.verbosity >= self.INFO

    @property
    def debug(self):
        """If verbosity is INFO or better"""
        return self.args.verbosity >= self.DEBUG

    def run(self):
        if not hasattr(self.args, 'command'):
            print("Use -h or --help for help")
            return

        method = getattr(self, "cmd_%s" % self.args.command, None)
        if self.debug:
            print("running %s" % method.__name__)
        if method is None:
            print("Command '%s' not implemented yet" % self.args.command.replace('_', ' '))
            return self.STATUS_CMD_NOT_FOUND
        try:
            return method()
        except exc.EgnyteError as e:
            if self.debug:
                raise
            print(repr(e))
            return self.STATUS_API_ERROR

    def get_client(self):
        result = client.EgnyteClient(self.config)
        if self.args.impersonate is not None:
            result.impersonate(self.args.impersonate)
        return result

    def get_access_token(self):
        config = self.require_password()
        return base.get_access_token(config)

    def merge_config(self):
        """Merge loaded config with command line params"""
        for key in self.config_keys:
            if getattr(self.args, key, None) is not None:
                self.config[key] = getattr(self.args, key)

    def require_password(self):
        """If config does not contain a password, ask user for it, but don't store it"""
        if self.config['password']:
            return self.config
        else:
            config = self.config.copy()
            config['password'] = getpass.getpass("Enter the password: ")
            return config

    def print_json(self, obj):
        print(json.dumps(obj, indent=2, sort_keys=True, default=to_json))

    def cmd_config_show(self):
        self.print_json(self.config)

    def cmd_config_create(self):
        self._config = {}
        self.merge_config()
        self.save_config()

    def cmd_config_update(self):
        self.merge_config()
        self.save_config()

    def cmd_config_token(self):
        self.config['access_token'] = self.get_access_token()
        self.save_config()

    def cmd_token(self):
        self.merge_config()
        print(self.get_access_token())

    def cmd_test(self):
        api = self.get_client()
        info = api.user_info()
        print("Connection successful for user %s" % (info['username'],))

    def cmd_search(self):
        api = self.get_client()
        results = api.search.files(self.args.query, modified_before=self.args.mtime_to, modified_after=self.args.mtime_from, folder=self.args.folder)
        self.print_json(results)


    def common_audit_args(self):
        format = self.args.format
        date_start = self.date(self.args.start)
        date_end = self.date(self.args.end)
        return (format, date_start, date_end)

    def date(self, value):
        """Poor mans human readable dates"""
        if value == 'today':
            return datetime.date.today()
        elif value == 'yesterday':
            return datetime.date.today() - datetime.timedelta(days=1)
        else:
            return datetime.date.datetime.strptime(value, "%Y-%m-%d").date()

    def wait_and_save_report(self, report):
        if self.args.save:
            output = open(self.args.save, "wb")
            print("Opened %s for writing, requesting report")
            with closing(output):
                report.wait()
                report.download().write_to(output)
        else:
            report.wait()
            download = report.download()
            with closing(download):
                lines = codecs.iterdecode(iter(download), 'UTF-8')
                for line in lines:
                    print(line)

    def comma_split(self, param):
        value = getattr(self.args, param, None)
        if value:
            return value.split(',')

    def cmd_audit_get(self):
        api = self.get_client()
        audits = api.audits
        report = audits.get(id=self.args.id)
        return self.wait_and_save_report(report)

    def cmd_audit_files(self):
        api = self.get_client()
        audits = api.audits
        folders = getattr(self.args, 'folder', None)
        file = self.args.file
        users = self.comma_split('users')
        transaction_type = self.comma_split('transaction_type')
        report = audits.files(*self.common_audit_args(), folders=folders, file=file, users=users, transaction_type=transaction_type)
        return self.wait_and_save_report(report)

    def cmd_audit_permissions(self):
        api = self.get_client()
        audits = api.audits
        assigners = self.comma_split('assigner')
        folders = self.args.folder
        users = self.comma_split('users')
        groups = self.comma_split('groups')
        report = audits.permissions(*self.common_audit_args(), assigners=assigners, folders=folders, users=users, groups=groups)
        return self.wait_and_save_report(report)

    def cmd_audit_logins(self):
        api = self.get_client()
        audits = api.audits
        users = self.comma_split('users')
        events = self.comma_split('events')
        access_points = self.comma_split('access_points')
        report = audits.logins(*self.common_audit_args(), events=events, access_points=access_points, users=users)
        return self.wait_and_save_report(report)

    def transfer_callbacks(self):
        if self.info:
            if sys.stdout.isatty():
                result = TerminalCallbacks()
                if self.debug:
                    result.force_newline = True
                return result
            else:
                return VerboseCallbacks()

    def cmd_upload(self):
        api = self.get_client()
        api.bulk_upload(self.args.paths, self.args.target, self.args.exclude, self.transfer_callbacks())

    def cmd_download(self):
        api = self.get_client()
        api.bulk_download(self.args.paths, self.args.target, self.args.overwrite, self.transfer_callbacks())

    def cmd_settings(self):
        self.print_json(self.get_client().settings)

    def cmd_events(self):
        start = self.args.start
        stop = self.args.stop
        events = self.get_client().events
        if start is None:
            start = events.latest_event_id
        elif start <= 0:
            start = events.latest_event_id + start
        if stop is not None and stop <= 0:
            stop = events.latest_event_id + stop
        events = events.filter(start_id = start, suppress=self.args.suppress, folder=self.args.folder, types=self.args.type or None)
        try:
            for event in events:
                self.print_json(event)
                print()
                if stop is not None and event.id >= stop:
                    break
        except KeyboardInterrupt:
            pass


class VerboseCallbacks(client.ProgressCallbacks):
    """Progress callbacks used when sys.stdout is a file or a pipe"""

    def write(self, text, force_newline=False):
        print(text)

    def getting_info(self, cloud_path):
        self.write("Getting info about %s" % cloud_path)

    def got_info(self, cloud_obj):
        self.write("Got info about %s" % cloud_obj.path)

    def download_start(self, local_path, cloud_file, size):
        self.write("Downloading %s" % local_path)
        self.current = local_path

    def upload_start(self, local_path, cloud_file, size):
        self.write("Uploading %s" % local_path)
        self.current = local_path

    def creating_directory(self, cloud_folder):
        self.write("Creating directory %s" % cloud_folder.path)

    def skipped(self, cloud_obj, reason):
        self.write("Skipped %s: %s" % (cloud_obj.path, reason), force_newline=True)

    def finished(self):
        self.write("Finished", force_newline=True)


class TerminalCallbacks(VerboseCallbacks):
    """Progress callbacks used when sys.stdout is a terminal"""
    force_newline = False

    def __init__(self):
        self.last_len = 0

    def write(self, text, force_newline=None):
        if force_newline is None:
            force_newline = self.force_newline
        output = ["\r"]
        sys.stdout.write("\r")  # return the carret
        if len(text) < self.last_len:  # clear out previous text
            sys.stdout.write(' ' * self.last_len)
            sys.stdout.write("\r")  # return the carret
        output.append(text)
        if force_newline:
            output.append('\n')
        sys.stdout.write("".join(output))
        sys.stdout.flush()
        self.last_len = len(text)

    def download_progress(self, cloud_file, size, downloaded):
        self.write("Downloading %s, %d%% complete" % (self.current, (downloaded * 100) / size))

    def upload_progress(self, cloud_file, size, uploaded):
        self.write("Uploading %s, %d%%" % (self.current, (uploaded * 100) / size))

    def download_finish(self, cloud_file):
        self.write("Downloaded %s" % self.current)

    def upload_finish(self, cloud_file):
        self.write("Uploaded %s" % self.current)


def main():
    parsed = create_main_parser().parse_args()
    sys.exit(Commands(parsed).run())

def full_help():
    parser = create_main_parser()
    return parser.format_help()

if __name__ == '__main__':
    main()



from egnyte.tests.config import EgnyteTestCase

FILE_PATH = '/search/test1.txt'
FILE_CONTENT = b'Lorem ipsum'
SEARCH_QUERY = u'ipsum'


class TestSearch(EgnyteTestCase):
    def setUp(self):
        super(TestSearch, self).setUp()
        self.root_folder.create()
        self.filepath = self.root_folder.path + FILE_PATH

    def test_file_search(self):
        _file = self.egnyte.file(self.filepath)
        _file.upload(FILE_CONTENT)
        search_results = self.egnyte.search.files(SEARCH_QUERY)
        # empty list is possible, as we won't get an answer here during the first run:
        # it takes ~1min for file to pass through indexing pipeline
        self.assertIsNotNone(search_results)
        if search_results:
            self.assert_(SEARCH_QUERY in search_results[0].snippet)



from egnyte import configuration, client, exc
import unittest

CONFIG_NAME = 'test_config.json'
ROOT_FOLDER_PATH = '/Shared/test_python_sdk/'


class EgnyteTestCase(unittest.TestCase):
    def setUp(self):
        self.config = configuration.load(CONFIG_NAME)
        self.egnyte = client.EgnyteClient(self.config)

        self.root_folder = self.egnyte.folder(ROOT_FOLDER_PATH)

    def tearDown(self):
        self.egnyte.folder(ROOT_FOLDER_PATH).delete()
        self.egnyte.close()
        del self.egnyte



from egnyte.tests.config import EgnyteTestCase

FOLDER_NAME = 'EVENT'


class TestEvents(EgnyteTestCase):
    def setUp(self):
        super(TestEvents, self).setUp()
        self.root_folder.create()

    def test_filter_poll(self):
        events = self.egnyte.events
        events = events.filter(events.oldest_event_id)
        results = events.poll(count=1)
        self.assertNotEqual(0, len(results), "Poll results should not be empty")
        self.assertNotEqual(events.start_id, events.oldest_event_id,
                            "latest_event_id should have been bumped after non-empty poll")

    def test_register_new_events(self):
        folder = self.root_folder.folder(FOLDER_NAME).create()

        events = self.egnyte.events
        events = events.filter(events.latest_event_id - 1)
        results = events.poll(count=1)

        self.assertEqual(results[0].action_source, 'PublicAPI')
        self.assertEqual(results[0].action, 'create')
        self.assertEqual(results[0].data['target_path'], folder.path)
        self.assertEqual(results[0].data['is_folder'], True)



from egnyte.tests.config import EgnyteTestCase
from egnyte import exc
from helpers import upload_file

FILE_FIRST_VERSION_NAME = 'FILE1.png'
FILE_SECOND_VERSION_NAME = 'FILE2.png'
TEXT_FILE_CONTENT = 'test_content'
DOWNLOADED_FILE_NAME = 'DOWNLOADED.png'
EGNYTE_FILE_NAME_IMAGE = '/sample.png'
EGNYTE_FILE_NAME_TEXT = '/test.txt'
DESTINATION_FOLDER_NAME = 'to_here'


class TestFiles(EgnyteTestCase):
    def setUp(self):
        super(TestFiles, self).setUp()
        self.filepath = self.root_folder.path + EGNYTE_FILE_NAME_IMAGE
        self.root_folder.create()

    def test_upload_file(self):
        uploaded_file = upload_file(self.egnyte, FILE_FIRST_VERSION_NAME, self.filepath)

        self.assertIsNone(uploaded_file.check())

    def test_download_file(self):
        uploaded_file = self.egnyte.file(self.root_folder.path + EGNYTE_FILE_NAME_TEXT)
        uploaded_file.upload(TEXT_FILE_CONTENT)

        downloaded_file = uploaded_file.download()

        self.assertEqual(downloaded_file.response.status_code, 200)
        self.assertEqual(downloaded_file.read(), TEXT_FILE_CONTENT)

    def test_copy_file(self):
        destination = self.root_folder.folder(DESTINATION_FOLDER_NAME)
        destination.create()
        uploaded_file = upload_file(self.egnyte, FILE_FIRST_VERSION_NAME, self.filepath)

        copied_file = uploaded_file.copy(destination.path + EGNYTE_FILE_NAME_IMAGE)

        self.assertIsNone(uploaded_file.check())
        self.assertIsNone(copied_file.check())

    def test_move_file(self):
        destination = self.root_folder.folder(DESTINATION_FOLDER_NAME)
        destination.create()
        uploaded_file = upload_file(self.egnyte, FILE_FIRST_VERSION_NAME, self.filepath)

        moved_file = uploaded_file.move(destination.path + EGNYTE_FILE_NAME_IMAGE)

        with self.assertRaises(exc.NotFound):
            uploaded_file.check()
        self.assertIsNone(moved_file.check())

    def test_delete_file(self):
        uploaded_file = upload_file(self.egnyte, FILE_FIRST_VERSION_NAME, self.filepath)

        self.assertIsNone(uploaded_file.check())

        uploaded_file.delete()

        with self.assertRaises(exc.NotFound):
            uploaded_file.check()

    def test_upload_new_version(self):
        upload_file(self.egnyte, FILE_FIRST_VERSION_NAME, self.filepath)
        second_version = upload_file(self.egnyte, FILE_SECOND_VERSION_NAME, self.filepath)
        file_attributes = second_version._fetch_attributes()

        self.assertEqual(file_attributes['num_versions'], 2)



"""
Exceptions and their handlers.
"""

from six.moves import http_client

__all__ = """
EgnyteError InvalidParameters InsufficientPermissions NotFound Redirected NotAuthorized JsonParseError DomainRequired ClientIdRequired
OAuthUsernameRequired OAuthPasswordRequired UnsupportedAuthStrategy RequestError DuplicateRecordExists FileSizeExceedsLimit ChecksumError
""".split()

class EgnyteError(Exception):
    """Base class for Egnyte SDK exceptions"""

    def __str__(self):
        """Pretty-printed version. Use repr for bare vesion instead"""
        contents = []
        for item in self:
            if isinstance(item, dict):
                contents.append("{%s}" % ", ".join(sorted(["%s: '%s'" % (k, v) for (k, v) in item.items()])))
            else:
                contents.append(str(item))
        return "<%s: %s>" % (self.__class__.__name__, ", ".join(contents))


class InvalidParameters(EgnyteError):
    """Invalid parameters were passed to an API request"""


class InsufficientPermissions(EgnyteError):
    """User does not have sufficient permissions to perform this action"""


class NotFound(EgnyteError):
    """Resource with name does not exist"""


class Redirected(EgnyteError):
    """Received unexpected HTTP 303 response"""


class NotAuthorized(EgnyteError):
    """Access token is required"""


class JsonParseError(EgnyteError):
    """Response from the server could not be parsed properly"""


class DomainRequired(EgnyteError):
    """Domain name is required"""


class ClientIdRequired(EgnyteError):
    """Client id is required"""


class OAuthUsernameRequired(EgnyteError):
    """Username is required for OAuth authentication"""


class OAuthPasswordRequired(EgnyteError):
    """Password is required for OAuth authentication"""


class UnsupportedAuthStrategy(EgnyteError):
    """This OAuth flow is not supported by this API key"""


class RequestError(EgnyteError):
    """Other kinds of request errors"""


class FileExpected(EgnyteError):
    """
    """


class DuplicateRecordExists(EgnyteError):
    """Existing entity conflict"""


class FileSizeExceedsLimit(EgnyteError):
    """File is too large for this operation."""


class ChecksumError(EgnyteError):
    """Checksum of the uploaded file is different than checksum calculated locally - file was corrupted during transfer."""

def extract_errors(data):
    """
    Try to extract useful information from inconsistent error data structures.
    """
    if 'errors' in data:
        data = data['errors']
    if 'inputErrors' in data:
        for err in extract_errors(data['inputErrors']):
            yield err
    elif hasattr(data, 'keys'):
        if 'code' in data:
            yield data
        else:
            for value in data.values():
                for err in extract_errors(value):
                    yield err
    elif isinstance(data, list):
        for value in data:
            for err in extract_errors(value):
                yield err
    else:
        yield data


def recursive_tuple(data):
    """Convert nested lists/dicts into tuples for structural comparing"""
    if isinstance(data, (list, tuple)):
        return tuple(recursive_tuple(x) for x in data)
    elif isinstance(data, dict):
        # We cannot use plain tuple(sorted(...)) here, because while it works giving us stable sort
        # in Python 2.7, it does not in Python 3 (results in TypeError: unorderable types
        # so we'll sort dictonaries first and then turn into tuples
        return tuple((recursive_tuple(x), recursive_tuple(y)) for (x, y) in sorted(data.items()))
    return data


class ErrorMapping(dict):
    """Maps HTTP status to EgnyteError subclasses"""
    ignored_errors = ()

    def __init__(self, values=None, ok_statuses=(http_client.OK, ), ignored_errors=None):
        super(ErrorMapping, self).__init__({
            http_client.BAD_REQUEST: RequestError,
            http_client.UNAUTHORIZED: NotAuthorized,
            http_client.FORBIDDEN: InsufficientPermissions,
            http_client.NOT_FOUND: NotFound,
            http_client.CONFLICT: DuplicateRecordExists,
            http_client.REQUEST_ENTITY_TOO_LARGE: FileSizeExceedsLimit,
            http_client.SEE_OTHER: Redirected,
        })
        if values:
            self.update(values)
        if ignored_errors:
            self.ignored_errors = recursive_tuple(ignored_errors)
        self.ok_statuses = ok_statuses

    def map_error(self, response):
        return self.get(response.status_code, RequestError)

    def check_response(self, response, *ok_statuses):
        """
        Check if HTTP response has a correct status,
        try to raise a specific EgnyteError subclass if not
        """
        if not len(ok_statuses):
            ok_statuses = self.ok_statuses
        if response.status_code not in ok_statuses:
            errors = [{'url': response.url}]
            error_type = self.map_error(response)
            try:
                data = response.json()
                for err in extract_errors(data):
                    errors.append(err)
            except Exception:
                errors.append({"http response": response.text})
            errors.append({"http status": response.status_code})
            errors.append({"headers": dict(response.headers)})
            if not self.ignore_error(errors):
                raise error_type(*errors)
        return response

    def ignore_error(self, errors):
        errors = recursive_tuple(errors[1:])
        if errors and self.ignored_errors:
            result = any((errors[:len(ignored)] == ignored) for ignored in self.ignored_errors)
            return result

    def check_json_response(self, response, *ok_statuses):
        """
        Check if HTTP response has a correct status and then parse it as JSON,
        try to raise a specific EgnyteError subclass if not
        """
        try:
            r = self.check_response(response, *ok_statuses)
            if r.status_code == http_client.NO_CONTENT:
                return None
            return r.json()
        except ValueError:
            raise JsonParseError({"http response": response.text})

    def copy(self):
        """Make a copy preserving class of self"""
        return self.__class__(self)

default = ErrorMapping()
partial = ErrorMapping(ok_statuses={http_client.PARTIAL_CONTENT})
accepted = ErrorMapping(ok_statuses={http_client.ACCEPTED})
created = ErrorMapping(ok_statuses={http_client.CREATED})
no_content_ok = ErrorMapping(ok_statuses={http_client.OK, http_client.NO_CONTENT})
created_ignore_existing = ErrorMapping(ok_statuses=(http_client.CREATED,), ignored_errors = [
    (u'Folder already exists at this location', {'http status': 403})
])



from __future__ import unicode_literals

import time

from egnyte import base, exc, resources



class Event(base.Resource):
    """
    Event data.

    Attributes:

    * id - event id
    * timestamp - date of event in iso8061 format
    * action_source - event source: WebUI, SyncEngine, Mobile or PublicAPI
    * actor - id of user that generate event
    * type - event type. For now we will support: file_system or note
    * action - event action. For now we will support: create,  delete,  move,  copy, or rename
    * object_detail - url to pub api that provide detail info about object from event like https://domain.egnyte.com/pubapi/v1/fs/Shared
    * data - additional data specific for event type and action

    Possible fields for 'data' field:

    for 'type'='file' and action create or delete

        'target_id' - entry id of create/deleted file
        'target_path' - path to created/deleted file

    for 'type'='file' and action move/copy/rename

        'source_path' - source path to moved/copied/renamed file
        'target_path' - target path to moved/copied/renamed file
        'source_id' - source entry id of moved/copied/renamed file (for move/rename there is one entry id so could be only one field or same data for source_id and target_id)
        'target_id' - target entry id of moved/copied/renamed file

    for 'type'='folder' and action create or delete

        'target_path' - path to created/deleted folder
        'folder_id' - folder id of created/deleted folder

    for 'type'='folder' and action move/copy/rename

        'source_path' - source path to moved/copied/renamed folder
        'target_path' - target path to moved/copied/renamed folder
        'source_id' - source folder id of moved/copied/renamed folder
        'target_id' - target folder id of moved/copied/renamed folder


    for 'type'='note' and any available action (create, delete)
        'note_id' - id of added/deleted note
    """
    _url_template = "pubapi/v1/events/%(id)s"

    def user(self):
        """Get a user object based on event attributes"""
        return resources.User(self._client, id=self.actor)


class Events(base.Resource):
    """
    Events.

    Attributes:

    * latest_event_id - id of latest event
    * oldest_event_id - id of oldest available event
    * timestamp - iso8601 timestamp of latest event
    """
    _url_template = "pubapi/v1/events/cursor"
    _url_template_list = "pubapi/v1/events"
    _lazy_attributes = {'latest_event_id', 'oldest_event_id', 'timestamp'}
    poll_delay = 1.0
    start_id = None
    suppress = None
    folder = None
    types = None

    def filter(self, start_id=None, suppress=None, folder=None, types=None):
        """
        Returns a filtered view of the events,

        Parameters:
        * start_id - return all events occurred after id from the previous request (the events shouldn't overlap between calls). defaults to latest_event_id
        * folder (optional) - return events occurred only for this folders and all its content (subfolders, files and notes).
        * suppress (optional) - filter out events from requesting client or filter out events from requesting client done by requesting user. Allowed values: app, user or none (defaults to no filter)
        * types (optional) - return only events of given types.

        """
        if types is not None:
            types = '|'.join(types)
        d = self.__dict__.copy()
        d.update(base.filter_none_values(dict(start_id=start_id, suppress=suppress, folder=folder, type=types)))
        return self.__class__(**d)

    def list(self, start_id, count=None):
        """
        Get detailed data about up to 'count' events 'start_id'.
        """
        if start_id is None:
            start_id = self.start_id
        params = base.filter_none_values(dict(id=start_id, suppress=self.suppress, type=self.types, count=count, folder=self.folder))
        url = self._client.get_url(self._url_template_list)
        json = exc.no_content_ok.check_json_response(self._client.GET(url, params=params))
        if json is None:
            return ()
        else:
            return base.ResultList((Event(self._client, **d) for d in json.get('events', ())), json['latest_id'], start_id)

    def poll(self, count=None):
        """
        List events starting with latest_event_id, if any found, update start_id and return them.
        """
        if self.start_id is None:
            self.start_id = self.latest_event_id
        results = self.list(self.start_id, count)
        if results:
            last = results[-1]
            self.start_id = last.id
            self.timestamp = last.timestamp
        return results

    def __iter__(self):
        """Never ending generator of events. Will block if necessary"""
        while True:
            results = self.poll()
            for x in results:
                yield x
            if not results:
                time.sleep(self.poll_delay)



import os


def upload_file(egnyte, filename, cloud_file_path):
    source = open(get_file_path(filename), 'rb')
    uploaded_file = egnyte.file(cloud_file_path)
    uploaded_file.upload(source)
    return uploaded_file


def get_file_path(filename):
    return os.path.join(os.getcwd(), 'egnyte', 'tests', 'data', filename)



from egnyte import configuration, client
import unittest

CONFIG_NAME = 'test_config.json'
GROUP_NAME = "Python SDK test group"
NEW_GROUP_NAME = "New name"


class TestGroups(unittest.TestCase):
    def setUp(self):
        self.config = configuration.load(CONFIG_NAME)
        self.egnyte = client.EgnyteClient(self.config)
        self.user = self.egnyte.users.by_username(self.config['login'])
        self.groups = self.egnyte.groups
        self.group = self.__createGroup(self.groups, GROUP_NAME, [self.user])

    def tearDown(self):
        self.group.delete()

    def test_list_groups(self):
        all_groups = self.egnyte.groups.list()

        self.assertGreaterEqual(all_groups, 1)
        self.assertIn(self.group, all_groups)

    def test_create_group(self):
        created_group = self.groups.by_displayName(GROUP_NAME)

        self.assertEqual(created_group, self.group)
        self.assertEqual(created_group.members[0]['username'], self.config['login'])

    def test_update_group(self):
        self.group.full_update(NEW_GROUP_NAME)

        updated_group = self.groups.by_displayName(NEW_GROUP_NAME)
        self.assertEqual(updated_group.members[0]['username'], self.config['login'])

    def __createGroup(self, groups, name,  members=None):
        return groups.create(name, members)



from __future__ import unicode_literals

import collections

import six

from egnyte import base, exc


class FileOrFolder(base.Resource):
    """Things that are common to both files and folders."""
    _url_template = "pubapi/v1/fs%(path)s"
    _lazy_attributes = {'name', 'folder_id', 'is_folder'}

    def _action(self, action, destination):
        exc.default.check_response(self._client.POST(self._url, dict(action=action, destination=destination)))
        return self.__class__(self._client, path=destination)

    def copy(self, destination):
        """Copy this to another path. Destination path should have all segments (including the last one)."""
        return self._action('copy', destination)

    def move(self, destination):
        """Move this to another path. Destination path should have all segments (including the last one)."""
        return self._action('move', destination)

    def link(self, accessibility, recipients=None, send_email=None, message=None,
             copy_me=None, notify=None, link_to_current=None,
             expiry_date=None, expiry_clicks=None, add_filename=None):
        """
        Create a link.

        * accessibility: Determines how the link can be accessed ('Anyone', 'Password', 'Domain', 'Recipients')
        * send_email: If true, Egnyte will send the link by email.
        * recipients: List email addresses for people who should receive the link. Only required if send_email is True (List of valid email addresses)
        * message: Personal message to be sent in link email. Only applies if send_email is True (plain text)
        * copy_me: If True, a copy of the link message will be sent to the link creator. Only applies if send_email is True.
        * notify: If True, link creator will be notified via email when link is accessed.
        * link_to_current: If True, link will always refer to current version of file. Only applicable for file links.
        * expiry_date: The expiry date for the link. If expiry_date is specified, expiry_clicks cannot be set (future date as datetime.date or string in YYYY-MM-DD format)
        * expiry_clicks: The number of times the link can be clicked before it stops working. If expiry_clicks is specified, expiry_date cannot be set (value must be between 1 - 10, inclusive)
        * add_filename: If True then the filename will be appended to the end of the link. Only applies to file links, not folder links.

        Will return sequence of created Links, one for each recipient.
        """
        return Links(self._client).create(path=self.path, type=self._link_kind, accessibility=accessibility,
                                          recipients=recipients, send_email=send_email, message=message,
                                          copy_me=copy_me, notify=notify, link_to_current=link_to_current,
                                          expiry_date=expiry_date, expiry_clicks=expiry_clicks, add_filename=add_filename)

    def _get(self):
        """Get the right object type (File or Folder), depending on what this path points to in the Cloud File System"""
        json = exc.default.check_json_response(self._client.GET(self._url))
        if json['is_folder'] and not isinstance(self, Folder):
            instance = Folder(self._client, path=self.path)
        elif not json['is_folder'] and not isinstance(self, File):
            instance = File(self._client, path=self.path)
        else:
            instance = self
        instance._update_attributes(json)
        if instance.is_folder:
            instance.folders = [Folder(self._client, **folder_data) for folder_data in json.get('folders', ())]
            instance.files = [File(self._client, **file_data) for file_data in json.get('files', ())]
        return instance


class File(FileOrFolder):
    """
    Wrapper for a file in the cloud.
    Does not have to exist - this can represent a new file to be uploaded.
    path - file path
    """
    _upload_chunk_size = 100 * (1024 * 1024)  # 100 MB
    _upload_retries = 3
    _link_kind = 'file'
    _lazy_attributes = {'num_versions', 'name', 'checksum', 'last_modified', 'entry_id',
                        'uploaded_by', 'size', 'is_folder'}
    _url_template_content = "pubapi/v1/fs-content%(path)s"
    _url_template_content_chunked = "pubapi/v1/fs-content-chunked%(path)s"

    def upload(self, fp, size=None, progress_callback=None):
        """
        Upload file contents.
        fp can be any file-like object, but if you don't specify it's size in advance it must support tell and seek methods.
        Progress callback is optional - if provided, it should match signature of ProgressCallbacks.upload_progress
        """
        if isinstance(fp, six.binary_type):
            fp = six.BytesIO(fp)
        if size is None:
            size = base.get_file_size(fp)
        if size < self._upload_chunk_size:
            # simple, one request upload
            retries = max(self._upload_retries, 1)
            while retries > 0:
                url = self._client.get_url(self._url_template_content, path=self.path)
                chunk = base._FileChunk(fp, 0, size)
                r = self._client.POST(url, data=chunk, headers={'Content-length': str(size)})
                exc.default.check_response(r)
                server_sha = r.headers['X-Sha512-Checksum']
                our_sha = chunk.sha.hexdigest()
                if server_sha == our_sha:
                    break
                retries -= 1
                # TODO: retry network errors too
            if retries == 0:
                raise exc.ChecksumError("Failed to upload file", {})
        else:  # chunked upload
            return self._chunked_upload(fp, size, progress_callback)

    def download(self, download_range=None):
        """
        Download file contents.
        Returns a FileDownload.
        Optional range is 2 integer sequence (start offset, end offset) used to download only part of the file.
        """
        url = self._client.get_url(self._url_template_content, path=self.path)
        if download_range is None:
            r = exc.default.check_response(self._client.GET(url, stream=True))
        else:
            if len(download_range) != 2:
                raise exc.InvalidParameters('Download range needs to be None or a 2 element integer sequence')
            r = exc.partial.check_response(self._client.GET(url, stream=True, headers={'Range': 'bytes=%d-%d' % download_range}))
        return base.FileDownload(r, self)

    def _chunked_upload(self, fp, size, progress_callback):
        url = self._client.get_url(self._url_template_content_chunked, path=self.path)
        chunks = list(base.split_file_into_chunks(fp, size, self._upload_chunk_size))  # need count of chunks
        chunk_count = len(chunks)
        headers = {}
        for chunk_number, chunk in enumerate(chunks, 1):  # count from 1 not 0
            headers['x-egnyte-chunk-num'] = "%d" % chunk_number
            headers['content-length'] = str(chunk.size)
            if chunk_number == chunk_count:  # last chunk
                headers['x-egnyte-last-chunk'] = "true"
            retries = max(self._upload_retries, 1)
            while retries > 0:
                r = self._client.POST(url, data=chunk, headers=headers)
                server_sha = r.headers['x-egnyte-chunk-sha512-checksum']
                our_sha = chunk.sha.hexdigest()
                if server_sha == our_sha:
                    break
                retries -= 1
                # TODO: retry network errors too
                # TODO: refactor common parts of chunked and standard upload
            if retries == 0:
                raise exc.ChecksumError("Failed to upload file chunk", {"chunk_number": chunk_number, "start_position": chunk.position})
            exc.default.check_response(r)
            if chunk_number == 1:
                headers['x-egnyte-upload-id'] = r.headers['x-egnyte-upload-id']
            if progress_callback is not None:
                progress_callback(self, size, chunk_number * self._upload_chunk_size)

    def delete(self):
        """Delete this file."""
        base.Resource.delete(self)

    def add_note(self, message):
        """Add a note to this file. Returns the created Note object."""
        return self._client.notes.create(self.path, message)

    def get_notes(self, **kwargs):
        """Get notes attached to this file. Returns list of Note objects"""
        return self._client.notes.list(file=self.path, **kwargs)


class Folder(FileOrFolder):
    """
    Wrapper for a folder the cloud.
    Does not have to exist - can represent a new folder yet to be created.
    """
    _url_template = "pubapi/v1/fs%(path)s"
    _url_template_permissions = "pubapi/v1/perms/folder/%(path)s"
    _url_template_effective_permissions = "pubapi/v1/perms/user/%(username)s"
    _lazy_attributes = {'name', 'folder_id', 'is_folder'}
    _link_kind = 'folder'
    folders = None
    files = None

    def folder(self, path, **kwargs):
        """Return a subfolder of this folder."""
        return Folder(self._client, path=self.path + '/' + path, **kwargs)

    def file(self, filename, **kwargs):
        """Return a file in this folder."""
        return File(self._client, folder=self, filename=filename, path=self.path + '/' + filename, **kwargs)

    def create(self, ignore_if_exists=True):
        """
        Create a new folder in the Egnyte cloud.
        If ignore_if_exists is True, error raised if folder already exists will be ignored.
        """
        r = self._client.POST(self._url, dict(action='add_folder'))
        (exc.created_ignore_existing if ignore_if_exists else exc.created).check_response(r)
        return self

    def delete(self):
        """Delete this folder in the cloud."""
        base.Resource.delete(self)

    def list(self):
        """
        Gets contents of this folder (in instance attributes 'folders' and 'files')
        """
        return self._get()

    def get_permissions(self, users=None, groups=None):
        """
        Get Permission values for this folder.
        """
        query_params = {}
        if users is not None:
            query_params[u'users'] = '|'.join(six.text_type(x) for x in users)
        if groups is not None:
            query_params[u'groups'] = '|'.join(six.text_type(x) for x in users)
        url = self._client.get_url(self._url_template_permissions, path=self.path)
        r = exc.default.check_json_response(self._client.GET(url, params=query_params))
        return PermissionSet(r)

    def set_permissions(self, permission, users=None, groups=None):
        """
        Set permission level for some users and/or groups for this folder.
        """
        url = self._client.get_url(self._url_template_permissions, path=self.path)
        data = base.filter_none_values(dict(permission=permission, users=users, groups=groups))
        exc.default.check_response(self._client.POST(url, data))

    def get_effective_permissions(self, username):
        """
        Get effective permissions (both direct, and granted by membership in groups) to this folder for a specific user.
        username: name of user (string)
        Returns one of 'Owner', 'Full', 'Editor', 'Viewer'
        """
        url = self._client.get_url(self._url_template_effective_permissions, username=username)
        params = dict(folder=self.path)
        r = exc.default.check_json_response(self._client.GET(url, params=params))
        return r['permission']

    def get_notes(self, **kwargs):
        """Get notes attached to any file in this folder."""
        return self._client.notes.list(folder=self.path, **kwargs)


class Link(base.Resource):
    """Link to a file or folder"""
    _url_template = "pubapi/v1/links/%(id)s"
    _lazy_attributes = {'copy_me', 'link_to_current', 'accessibility', 'notify',
                        'path', 'creation_date', 'type', 'send_mail'}

    def delete(self):
        """Delete this link"""
        base.Resource.delete(self)


class User(base.Resource):
    """
    Wrapper for a User.
    Warning: attribute names in this class use camelCase instead of underscores.
    Name is a dictionary with 2 keys: givenName and lastName.
    """
    _url_template = "pubapi/v2/users/%(id)s"
    _url_template_effective_permissions = "pubabi/v1/perms/user/%(userName)s"
    _lazy_attributes = {'userName', 'externalId', 'email', 'name', 'active', 'locked', 'authType',
                        'role', 'userType', 'idpUserId'}

    def delete(self):
        """Delete this user account."""
        base.Resource.delete(self)

    def update(self, email=None, familyName=None, givenName=None, active=None, sendInvite=None, authType=None,
               userType=None, idpUserId=None, userPrincipalName=None):
        """
        Modify this user account.
        Optional parameters (no change if value is None):

        * email: The email address of the user. Any valid email address (e.g. admin@acme.com)
        * familyName: The last name of the user. Any plain text (e.g. John)
        * givenName: The first name of the user. Any plain text (e.g. Smith)
        * active: Whether the user is active or inactive. True or False
        * sendInvite: If set to true when creating a user, an invitation email will be sent (if the user is created in active state). True or False
        * authType: The authentication type for the user. 'ad' (AD), 'sso' (SAML SSO), 'egnyte' (Internal Egnyte)
        * userType: The Egnyte role of the user. 'admin' (Administrator), 'power' (Power User), 'standard' (Standard User)
        * idpUserId: Only required if the user is SSO authenticated and not using default user mapping. Do not specify if user is not SSO authenticated. This is the way the user is identified within the SAML Response from an SSO Identity Provider, i.e. the SAML Subject (e.g. jsmith)
        * userPrincipalName: Do not specify if user is not AD authenticated. Used to bind child authentication policies to a user when using Active Directory authentication in a multi-domain setup (e.g. jmiller@example.com)
        """
        url = self._client.get_url(self._url_template, id=self.id)
        name = base.filter_none_values(dict(familyName=familyName, givenName=givenName)) or None
        data = base.filter_none_values(dict(email=email, active=active, name=name, sendInvite=sendInvite,
                                            authType=authType, userType=userType, idpUserId=idpUserId,
                                            userPrincipalName=userPrincipalName))
        json = exc.default.check_json_response(self._client.PATCH(url, data))
        self._update_attributes(json)

    def get_effective_permissions(self, path):
        url = self._client.get_url(self._url_template_effective_permissions, userName=self.userName)
        r = exc.default.check_json_response(self._client.GET(url, params=dict(folder=path)))
        return r


class Note(base.Resource):
    """Note attached to a file or folder."""
    _url_template = "pubapi/v1/notes/%(id)s"
    _lazy_attributes = {'file_id', 'file_path', 'can_delete', 'creation_time', 'message', 'username', 'formatted_name'}

    def delete(self):
        """Delete this Note"""
        base.Resource.delete(self)

    def get_file(self):
        """Get the file to which this note is attached."""
        return self._client.file(self.file_path)

class Group(base.Resource):
    """Group of users."""
    _url_template = "pubapi/v2/groups/%(id)s"
    _lazy_attributes = {'displayName', 'members'}

    def delete(self):
        """Delete this Group"""
        base.Resource.delete(self)

    def full_update(self, displayName, members=None):
        """
        Full update of this group.

        This endpoint is used to overwrite all of the attributes of a group. This is especially useful for making a change to settings that ensures all prior settings are removed.

        * displayName: Name of the group (string). Required
        * members: List of members to be added to the new group (user ids or User objects). Optional.
        """
        url = self._client.get_url(self._url_template, id=self.id)
        data = dict(displayName=displayName)
        if members is not None:
            data['members'] = [dict(value=x.id if isinstance(x, User) else x) for x in members]
        json = exc.default.check_json_response(self._client.PATCH(url, data))
        self._update_attributes(json)


class Links(base.HasClient):
    """Link management API"""
    _url_template = "pubapi/v1/links"

    def create(self, path, type, accessibility,
               recipients=None, send_email=None, message=None,
               copy_me=None, notify=None, link_to_current=None,
               expiry_date=None, expiry_clicks=None, add_filename=None,
               ):
        """
        Create links.

        * path:  The absolute path of the destination file or folder.
        * type:  This determines what type of link will be created ('File' or 'Folder')
        * accessibility: Determines who a link is accessible by ('Anyone', 'Password', 'Domain', 'Recipients')
        * send_email: If True, the link will be sent via email by Egnyte.
        * recipients: List email addresses of recipients of the link. Only required if send_email is True (List of valid email addresses)
        * message: Personal message to be sent in link email. Only applies if send_email is True (plain text)
        * copy_me: If True, a copy of the link message will be sent to the link creator. Only applies if send_email is True.
        * notify: If True, link creator will be notified via email when link is accessed.
        * link_to_current: If True, link will always refer to current version of file. Only applicable for file links.
        * expiry_date: The expiry date for the link. If expiry_date is specified, expiry_clicks cannot be set (future date as datetime.date or string in YYYY-MM-DD format)
        * expiry_clicks: The number of clicks the link is valid for. If expiry_clicks is specified, expiry_date cannot be set (value must be between 1 - 10, inclusive)
        * add_filename: If True then the filename will be appended to the end of the link. Only applies to file links, not folder links.

        Will return a sequence of created Links, one for each recipient.
        """
        url = self._client.get_url(self._url_template)
        data = base.filter_none_values(dict(path=path, type=type, accessibility=accessibility, send_email=send_email,
                                            copy_me=copy_me, notify=notify, add_filename=add_filename, link_to_current=link_to_current,
                                            expiry_clicks=expiry_clicks, expiry_date=base.date_format(expiry_date),
                                            recipients=recipients, message=message))
        response = exc.default.check_json_response(self._client.POST(url, data))
        # This response has weird structure
        links = response.pop('links')
        result = []
        for l in links:
            l.update(response)
            result.append(Link(self._client, **l))
        return result

    def get(self, id):
        """Get a Link object by it's id"""
        return Link(self._client, id=id)

    def list(self, path=None, username=None, created_before=None, created_after=None, type=None, accessibility=None,
             offset=None, count=None):
        """
        Search links that match following optional conditions:

        * path: List links to this file or folder (Full absolute path of destination file or folder)
        * username: List links created by this user (Any username from your Egnyte account)
        * created_before: List links created before this date (datetime.date, or string in YYYY-MM-DD format)
        * created_after: List links created after this date (datetime.date, or string in YYYY-MM-DD format)
        * type: Links of selected type will be shown ('File' or 'Folder')
        * accessibility: Links of selected accessibility will be shown ('Anyone', 'Password', 'Domain', or 'Recipients')
        * offset: Start at this link, where offset=0 means start with first link.
        * count: Send this number of links. If not specified, all links will be sent.

        Returns a list of Link objects, with additional total_count and offset attributes.
        """
        url = self._client.get_url(self._url_template)
        params = base.filter_none_values(dict(path=path, username=username, created_before=base.date_format(created_before),
                                              created_after=base.date_format(created_after), type=type, accessibility=accessibility,
                                              offset=offset, count=count))
        json = exc.default.check_json_response(self._client.GET(url, params=params))
        return base.ResultList((Link(self._client, id=id) for id in json.get('ids', ())), json['total_count'], json['offset'])


class Users(base.HasClient):
    """User management API"""
    _url_template = "pubapi/v2/users"

    def list(self, email=None, externalId=None, userName=None, startIndex=None, count=None):
        """
        Search users. Optional search parameters are 'email', 'externalId' and 'userName'.
        startIndex (starts with 1) and count may be used for pagination

        Returns a list of User objects, with additional total_count and offset attributes.
        """
        url = self._client.get_url(self._url_template)
        filters = base.filter_none_values(dict(email=email, externalId=externalId, userName=userName))
        params = base.filter_none_values(dict(startIndex=startIndex, count=count))
        params['filter'] = [u'%s eq "%s"' % (k, v) for (k, v) in filters.items()]
        json = exc.default.check_json_response(self._client.GET(url, params=params))
        return base.ResultList((User(self._client, **d) for d in json.get('resources', ())), json['totalResults'], json['startIndex'] - 1)

    def get(self, id):
        """Get a User object by id. Does not check if User exists."""
        return User(self._client, id=id)

    def by_email(self, email):
        """Get a User object by email. Returns None if user does not exist"""
        try:
            return self.list(email=email)[0]
        except LookupError:
            pass

    def by_username(self, userName):
        """Get a User object by username. Returns None if user does not exist"""
        try:
            return self.list(userName=userName)[0]
        except LookupError:
            pass

    def create(self, userName, externalId, email, familyName, givenName, active=True, sendInvite=True, authType='egnyte',
               userType='power', role=None, idpUserId=None, userPrincipalName=None):
        """
        Create a new user account. Parameters:

        * userName: The Egnyte username for the user. Username must start with a letter or digit. Special characters are not supported (with the exception of periods, hyphens, and underscores).
        * externalId: This is an immutable unique identifier provided by the API consumer. Any plain text (e.g. S-1-5-21-3623811015-3361044348-30300820-1013)
        * email: The email address of the user. Any valid email address (e.g. admin@acme.com)
        * familyName: The last name of the user. Any plain text (e.g. John)
        * givenName: The first name of the user. Any plain text (e.g. Smith)
        * active: Whether the user is active or inactive. True or False
        * sendInvite: If set to true when creating a user, an invitation email will be sent (if the user is created in active state). True or False
        * authType: The authentication type for the user. 'ad' (AD), 'sso' (SAML SSO), 'egnyte' (Internal Egnyte)
        * userType: The type of the user. 'admin' (Administrator), 'power' (Power User), 'standard' (Standard User)
        * role: The role assigned to the user. Only applicable for Power Users. Default or custom role name
        * idpUserId: Only required if the user is SSO authenticated and not using default user mapping. Do not specify if user is not SSO authenticated. This is the way the user is identified within the SAML Response from an SSO Identity Provider, i.e. the SAML Subject (e.g. jsmith)
        * userPrincipalName: Do not specify if user is not AD authenticated. Used to bind child authentication policies to a user when using Active Directory authentication in a multi-domain setup (e.g. jmiller@example.com)

        Returns created User object.
        """
        url = self._client.get_url(self._url_template)
        data = base.filter_none_values(dict(userName=userName, externalId=externalId, email=email,
                                            name=dict(familyName=familyName, givenName=givenName), active=active, sendInvite=sendInvite, authType=authType,
                                            userType=userType, role=role, idpUserId=idpUserId, userPrincipalName=userPrincipalName))
        json = exc.created.check_json_response(self._client.POST(url, data))
        return User(self._client, **json)


class PermissionSet(object):
    """Wrapper for a permission set"""

    def __init__(self, json):
        self._users = json.get('users', ())
        self._groups = json.get('groups', ())
        self._unpack()

    def _unpack(self):
        self.user_to_permission = {}
        self.group_to_permission = {}
        self.permission_to_owner = collections.defaultdict(lambda: dict(users=set(), groups=set()))
        for d in self._users:
            self.user_to_permission[d['subject']] = d['permission']
            self.permission_to_owner[d['permission']]['users'].add(d['subject'])
        for d in self._groups:
            self.group_to_permission[d['subject']] = d['permission']
            self.permission_to_owner[d['permission']]['groups'].add(d['subject'])


class Notes(base.HasClient):
    """
    Notes management API
    """
    _url_template = "pubapi/v1/notes"

    def create(self, path, message):
        """
        Create a new note.
        Parameters:

        * path - path to the file the note is about
        * message - contents of the note

        Returns the created Note object.
        """
        url = self._client.get_url(self._url_template)
        data = dict(path=path, body=message)
        json = exc.created.check_json_response(self._client.POST(url, data))
        return Note(self._client, **json)

    def list(self, file=None, folder=None, start_time=None, end_time=None):
        """
        List existing notes.
        Optional filtering parameters:

        * start_time: Get notes created after start_time (datetime.date or string in 'YYYY-MM-DD' format)
        * file: Get only notes attached to a specific file (path).
        * folder: Get only notes atatched to files in specific folder (path).
        * end_time: Get notes created before end_time (datetime.date or string in 'YYYY-MM-DD' format)

        Returns list of Note objects, with additional attributes total_count and offset.
        """
        url = self._client.get_url(self._url_template)
        params = base.filter_none_values(dict(file=file, folder=folder, start_time=base.date_format(start_time),
                                              end_time=base.date_format(end_time)))
        json = exc.default.check_json_response(self._client.GET(url, params=params))
        return base.ResultList((Note(self._client, **d) for d in json.pop('notes', ())), json['total_results'], json['offset'])


class Groups(base.HasClient):
    """
    Group Management API
    """
    _url_template = "pubapi/v2/groups"

    def list(self, displayName=None, startIndex=None, count=None):
        """
        List existing groups.
        Optional filtering parameters:
        
        * displayName: Filter by name of the group. This may contain '*' wildcards at beginning for prefix search or both at beginning and end for contains search.

        Returns list of Group objects, with additional attributes total_result and offset
        """
        params = base.filter_none_values(dict(startIndex=startIndex, count=count))
        if displayName:
            if displayName.startswith('*'):
                op = 'co' if displayName.endswith('*') else 'sw'
            else:
                op = 'eq'
            params['filter'] = [u'displayName %s "%s"' % (op, displayName.strip('*'))]
        url = self._client.get_url(self._url_template)
        json = exc.default.check_json_response(self._client.GET(url, params=params))
        return base.ResultList((Group(self._client, **d) for d in json.pop('resources', ())), json['totalResults'], json['startIndex'] - 1)

    def create(self, displayName, members=None):
        """
        Create a new Group. Parameters:

        * displayName: Name of the group (string). Required
        * members: List of members to be added to the new group (user ids or User objects). Optional.

        Returns created Group object.
        """
        url = self._client.get_url(self._url_template)
        data = dict(displayName=displayName)
        if members is not None:
            data['members'] = [dict(value=x.id if isinstance(x, User) else x) for x in members]
        json = exc.created.check_json_response(self._client.POST(url, data))
        return Group(self._client, **json)
    

    def get(self, id):
        """Get a Group object by id. Does not check if Group exists."""
        return Group(self._client, id=id)

    def by_displayName(self, displayName):
        """Get a Group object by displayName. Returns None if Group does not exist"""
        try:
            return self.list(displayName=displayName)[0]
        except LookupError:
            pass

class SearchMatch(base.HasClient):
    """
    Single match from search results.
    Attributes for a file match:
    
    * name The name of the file.
    * path The path to the file in Egnyte.
    * type The MIME type of the file.
    * size The size of the file in bytes.
    * snippet A plain text snippet of the text containing the matched content.
    * snippet_html An HTML formatted snippet of the text containing the matched content.
    * entry_id A GUID for tha particular instance of a file.
    * last_modified The ISO-8601 formatted timestamp representing the last modified date of the file.
    * uploaded_by The formatted name of the user who uploaded the file.
    * uploaded_by_username The username of the user who uploaded the file.
    * num_versions The number of versions of the file available.
    * is_folder A boolean value stating if the object is a file or folder. Please note that, currently, this API only returns file objects.
    """

    def file(self):
        """Get File object that correspons to this search match, or None if found resource is not a File"""
        if not self.is_folder:
            return File(self._client, name=self.name, path=self.path, is_folder=self.is_folder, num_versions=self.num_versions,
                        entry_id=self.entry_id, uploaded_by=self.uploaded_by, size=self.size, last_modified=self.last_modified)


class Search(base.HasClient):
    """Search API"""
    _url_template = "pubapi/v1/search"

    def files(self, query, offset=None, count=None, folder=None, modified_after=None, modified_before=None):
        """
        Search for files.
        Parameters:

        * query The search string you want to find. * is supported as a postfix wildcard, AND and OR as bool operations and double quotes for phrase search.
        * offset The 0-based index of the initial record being requested (Integer >= 0).
        * count The number of entries per page (min 1, max 100)
        * folder Limit the result set to only items contained in the specified folder.
        * modified_before Limit to results before the specified ISO-8601 timestamp (datetime.date object or string).
        * modified_after Limit to results after the specified ISO-8601 timestamp (datetime.date object or string).

        Returns list of SearchMatch objects, with additional attributes total_count and offset.
        """
        url = self._client.get_url(self._url_template)
        params = base.filter_none_values(dict(
            query=query,
            offset=offset,
            count=count,
            folder=folder,
            modified_after=base.date_format(modified_after),
            modified_before=base.date_format(modified_before))
        )
        json = exc.default.check_json_response(self._client.GET(url, params=params))
        return base.ResultList((SearchMatch(self._client, **d) for d in json.get('results', ())), json['total_count'], json['offset'])



from __future__ import print_function

import collections
import os
import os.path
import shutil

from egnyte import exc, base, resources, audits, events


class EgnyteClient(base.Session):
    """Main client objects. This should be the only object you have to manually create in standard API use."""
    @property
    def links(self):
        """API for Links management"""
        return resources.Links(self)

    @property
    def user_info(self):
        """
        Information about user associated with this API access token.
        """
        return exc.default.check_json_response(self.GET(self.get_url("pubapi/v1/userinfo")))

    @property
    def users(self):
        """API for User management"""
        return resources.Users(self)

    @property
    def audits(self):
        """API for Audit Reports"""
        return audits.Audits(self)

    @property
    def notes(self):
        """API for Notes management"""
        return resources.Notes(self)

    @property
    def groups(self):
        """API for Group Management"""
        return resources.Groups(self)

    @property
    def search(self):
        """API for Search"""
        return resources.Search(self)

    @property
    def events(self):
        """API for events"""
        return events.Events(self)

    @property
    def settings(self):
        """Domain settings."""
        if not hasattr(self, '_cached_settings'):
            result  = {}
            urls = (
                ('users', "pubapi/v2/users/settings"),
                ('links', "pubapi/v1/links/settings"),
                ('file_system', "pubapi/v1/fs/settings"),
                ('audit', "pubapi/v1/audit/settings"),
            )
            for name, url in urls:
                try:
                    result[name] = exc.default.check_json_response(self.GET(self.get_url(url)))
                except Exception: #exc.InsufficientPermissions:
                    pass

            self._cached_settings = result
        return self._cached_settings

    def folder(self, path="/Shared", **kwargs):
        """Get a Folder object for the specified path"""
        return resources.Folder(self, path=path.rstrip('/'), **kwargs)

    def file(self, path, **kwargs):
        """Get a File object for the specified path"""
        return resources.File(self, path=path, **kwargs)

    def get(self, path):
        """Check whether a path is a file or a folder and return the right object."""
        return self.folder(path)._get()

    def impersonate(self, username):
        """
        Start impersonating another user.

        * username: either username or full email address of user to impersonate.
        """
        self._session.headers['X-Egnyte-Act-As-Email' if '@' in username else 'X-Egnyte-Act-As'] = username

    def stop_impersonating(self):
        """
        Stop impersonating another user.
        """
        self._session.headers.pop('X-Egnyte-Act-As', None)
        self._session.headers.pop('X-Egnyte-Act-As-Email', None)

    def bulk_upload(self, paths, target, exclude=None, progress_callbacks=None):
        """
        Transfer many files or directories to Cloud File System.

        * paths - list of local file paths
        * target - Path in CFS to upload to
        * progress_callbacks - Callback object (see ProgressCallbacks)
        """
        if not paths:
            return
        if progress_callbacks is None:
            progress_callbacks = ProgressCallbacks()  # no-op callbacks
        target_folder = self.folder(target)
        progress_callbacks.creating_directory(target_folder)
        target_folder.create(True)
        for is_dir, local_path, cloud_path in base.generate_paths(paths, exclude):
            if is_dir:
                cloud_dir = target_folder.folder(cloud_path)
                progress_callbacks.creating_directory(cloud_dir)
                cloud_dir.create(True)
            else:
                size = os.path.getsize(local_path)
                if size:  # empty files cannot be uploaded
                    cloud_file = target_folder.file(cloud_path, size=size)
                    with open(local_path, "rb") as fp:
                        progress_callbacks.upload_start(local_path, cloud_file, size)
                        cloud_file.upload(fp, size, progress_callbacks.upload_progress)
                    progress_callbacks.upload_finish(cloud_file)
        progress_callbacks.finished()

    def _bulk_download(self, items, root_path, local_dir, overwrite, progress_callbacks):
        root_len = len(root_path.rstrip('/')) + 1
        queue = collections.deque(items)
        while True:
            try:
                obj = queue.popleft()
            except IndexError:
                break
            relpath = obj.path[root_len:].strip('/')
            local_path = os.path.join(local_dir, relpath.replace('/', os.sep))
            dir_path = os.path.dirname(local_path)
            if not os.path.isdir(dir_path):
                if os.path.exists(dir_path):
                    if overwrite:
                        os.unlink(local_path)
                    else:
                        progress_callbacks.skipped(obj, "Existing file conflicts with cloud folder")
                        continue
                os.makedirs(dir_path)
            if obj.is_folder:
                # schedule contents for later, files first
                if obj.files is None:
                    progress_callbacks.getting_info(obj.path)
                    obj.list()
                    progress_callbacks.got_info(obj)
                queue.extend(obj.files)
                queue.extend(obj.folders)
            else:
                if os.path.exists(local_path):
                    if overwrite:
                        if os.path.isdir(local_path) and not os.path.islink(local_path):
                            shutil.rmtree(local_path)
                        else:
                            os.unlink(local_path)
                    else:
                        progress_callbacks.skipped(obj, "Existing file conflicts with cloud file")
                        continue
                progress_callbacks.download_start(local_path, obj, obj.size)
                obj.download().save_to(local_path, progress_callbacks.download_progress)
                progress_callbacks.download_finish(obj)

    def bulk_download(self, paths, local_dir, overwrite=False, progress_callbacks=None):
        """
        Transfer many files or directories to Cloud File System.

        * paths - list of local file paths
        * target - Path in CFS to upload to
        * progress_callbacks - Callback object (see ProgressCallbacks)
        """
        if progress_callbacks is None:
            progress_callbacks = ProgressCallbacks()
        for path in paths:
            progress_callbacks.getting_info(path)
            obj = self.get(path)
            progress_callbacks.got_info(obj)
            root_path = path[:path.rstrip('/').rfind('/')]  # take all segments expect last one
            if obj.is_folder:
                items = obj.files + obj.folders
            else:
                items = (obj,)
            self._bulk_download(items, root_path, local_dir, overwrite, progress_callbacks)
        progress_callbacks.finished()


class ProgressCallbacks(object):
    """
    This object is used for bulk transfers (uploads and downloads)
    Inherit this and add override any of the callabcks you'd like to handle.
    """

    def getting_info(self, cloud_path):
        """Getting information about an object. Called for directories and unknown paths."""

    def got_info(self, cloud_obj):
        """Got information about an object."""

    def creating_directory(self, cloud_folder):
        """Creating a directory."""

    def download_start(self, local_path, cloud_file, size):
        """Starting to download a file."""

    def download_progress(self, cloud_file, size, downloaded):
        """Some progress in file download."""

    def download_finish(self, cloud_file):
        """Finished downloading a file."""

    def upload_start(self, local_path, cloud_file, size):
        """Starting to upload a file."""

    def upload_progress(self, cloud_file, size, uploaded):
        """Some progress in file upload."""

    def upload_finish(self, cloud_file):
        """Finished uploading a file."""

    def finished(self):
        """Called after all operations."""

    def skipped(self, cloud_obj, reason):
        """Object has been skipped because of 'reason'"""



from egnyte import configuration, client
import unittest

CONFIG_NAME = 'test_config.json'
USERNAME = "python_sdk_test_user"
EXTERNAL_ID = "python_sdk_test_user"
EMAIL = "python_sdk_test_user@example.com"
ANOTHER_EMAIL = "python_sdk_another_email@example.com"
FAMILY_NAME = "Doe"
GIVEN_NAME = "John"
ACTIVE = False
SEND_INVITE = False


class TestUserInfo(unittest.TestCase):
    def test_userinfo(self):
        self.config = configuration.load(CONFIG_NAME)
        self.egnyte = client.EgnyteClient(self.config)
        data = self.egnyte.user_info
        self.assertEqual(data["username"], self.config['login'],
                         "Username received from API does not match one in config file")


class TestUsers(unittest.TestCase):
    def setUp(self):
        self.config = configuration.load(CONFIG_NAME)
        self.egnyte = client.EgnyteClient(self.config)
        self.users = self.egnyte.users
        self.user = self.__create_user(self.users)

    def tearDown(self):
        self.user.delete()

    def test_list_users(self):
        all_users = self.egnyte.users.list()
        self.assertGreaterEqual(all_users, 1)

    def test_create_user(self):
        user_by_email = self.users.by_email(EMAIL)
        self.assertEqual(self.user, user_by_email, "Should find user by email")

        user_by_username = self.users.by_username(USERNAME)
        self.assertEqual(self.user, user_by_username, "Should find user by username")

    def test_update_user(self):
        self.user.update(email=ANOTHER_EMAIL)

        user_updated = self.users.by_email(EMAIL)
        self.assertIsNone(user_updated, "Should not find user after email has changed")

        user_updated = self.users.by_email(ANOTHER_EMAIL)
        self.assertEqual(self.user, user_updated, "Should find user by new email")

    def __create_user(self, users):
        return users.create(userName=USERNAME,
                            externalId=EXTERNAL_ID,
                            email=EMAIL,
                            familyName=FAMILY_NAME,
                            givenName=GIVEN_NAME,
                            active=False,
                            sendInvite=False)



from __future__ import print_function, unicode_literals

import datetime
import fnmatch
import hashlib
import json
import os
import os.path
import re
import time
import requests

from six import string_types, text_type
from six.moves.urllib.parse import quote
from egnyte import exc, configuration


JSON_HEADERS = {'content-type': 'application/json'}
DEFAULT_TIMEOUT = 30


class Session(object):
    """
    Provides persistent HTTPS connections to the Egnyte API
    """
    time_between_requests = None
    last_request_time = None

    def __init__(self, config=None):
        self.config = config if isinstance(config, dict) else configuration.load(config)
        domain = self.config['domain']
        if '.' not in domain:
            domain += ".egnyte.com"
        self._url_prefix = "https://%s/" % domain
        self._session = requests.Session()
        if 'access_token' in self.config:
            self._session.headers['Authorization'] = 'Bearer %s' % self.config['access_token']
        if 'time_between_requests' in self.config:
            self.time_between_requests = config['time_between_requests']
        elif 'requests_per_second' in self.config:
            self.time_between_requests = 1.0 / float(self.config['requests_per_second'])

    def _respect_limits(self):
        if self.time_between_requests:
            if self.last_request_time is not None:
                since = time.time() - self.last_request_time
                if since < self.time_between_requests:
                    time.sleep(self.time_between_requests - since)
            self.last_request_time = time.time()

    def _retry(self, func, *args, **kwargs):

        kwargs["timeout"] = int(self.config.get("timeout")) if self.config.get(
            "timeout") is not None else DEFAULT_TIMEOUT

        while True:
            response = func(*args, **kwargs)
            if response.headers.get('x-mashery-error-code') == 'ERR_403_DEVELOPER_OVER_QPS':
                retry_after = float(response.headers.get('retry-after', '1'))
                time.sleep(retry_after)
            else:
                return response

    def GET(self, url, **kwargs):
        self._respect_limits()
        return self._retry(self._session.get, url, allow_redirects=False, **kwargs)

    def POST(self, url, json_data=None, **kwargs):
        self._respect_limits()
        if json_data is None:
            headers = {}
            data = kwargs.pop('data', None)
        else:
            headers = JSON_HEADERS
            data = json.dumps(json_data)
        if 'headers' in kwargs:
            headers.update(kwargs.pop('headers'))
        return self._retry(self._session.post, url, data=data, headers=headers, **kwargs)

    def PATCH(self, url, json_data=None, **kwargs):
        self._respect_limits()
        if json_data is None:
            headers = {}
            data = kwargs.pop('data', None)
        else:
            headers = JSON_HEADERS
            data = json.dumps(json_data)
        if 'headers' in kwargs:
            headers.update(kwargs.pop('headers'))
        return self._retry(self._session.patch, url, data=data, headers=headers, **kwargs)

    def DELETE(self, url, **kwargs):
        self._respect_limits()
        return self._retry(self._session.delete, url, **kwargs)

    def get_url(self, _path, **kwargs):
        if kwargs:
            kw = {k: encode_path(v) if isinstance(v, string_types) else str(v) for k, v in kwargs.items()}
            return self._url_prefix + _path % kw
        else:
            return self._url_prefix + _path

    def close(self):
        if hasattr(self, '_session'):
            self._session.close()
            del self._session


class HasClient(object):
    """Base class for API wrappers and utils"""

    def __init__(self, _client, **kwargs):
        self._client = _client
        self.__dict__.update(kwargs)


class Resource(object):
    """Base wrapper for API resources (singular objects with specific URL)"""
    _lazy_attributes = ()
    _url_template = ""  # Whatever this depends on should not be in _lazy_attributes

    def __init__(self, _client, **kwargs):
        self._client = _client
        self.__dict__.update(kwargs)
        if '_url' not in kwargs:
            self._url = self._client.get_url(self._url_template, **kwargs)

    def __getattr__(self, name):
        """If attribute is in _lazyAtrributes but we don't have it's value yet, fetch attributes from service."""
        if name in self._lazy_attributes:
            if name not in self.__dict__:
                self._fetch_attributes()
            if name in self.__dict__:
                return self.__dict__[name]
        raise AttributeError(self, name)

    def _update_attributes(self, json_dict):
        for key in self._lazy_attributes:
            if key in json_dict:
                self.__dict__[key] = json_dict[key]

    def _fetch_attributes(self):
        json = exc.default.check_json_response(self._client.GET(self._url))
        self._update_attributes(json)
        return json

    def check(self):
        """
        Check if this object exists in the cloud and current user has read permissions on it.
        Will raise an exception otherwise.
        """
        self._fetch_attributes()

    def __str__(self):
        return "<%s: %s {%s} >" % (self.__class__.__name__, self._url, ", ".join(["%s: %r" % (k, v) for (k, v) in sorted(self.__dict__.items()) if not k.startswith('_')]))

    __repr__ = __str__

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        if isinstance(other, Resource):
            return (self._client is other._client and self._url == other._url)
        return NotImplemented

    def delete(self):
        exc.default.check_response(self._client.DELETE(self._url))


def get_access_token(config):
    session = Session(config)
    url = session.get_url("puboauth/token")
    data = dict(
        client_id=config['api_key'],
        username=config['login'],
        password=config['password'],
        grant_type="password",
    )
    response = session.POST(url, data, headers={'content-type': 'application/x-www-form-urlencoded'})
    return exc.default.check_json_response(response)['access_token']


class _FileChunk(object):
    """Wrapper for chunk of the file that also calculates SHA512 checksum while file is read"""

    def __init__(self, fp, start, size):
        self.fp = fp
        self.position = start
        self.left = self.size = size
        self.sha = hashlib.sha512()

    def read(self, size=None):
        if size is None or size > self.left:
            size = self.left
        result = self.fp.read(size)
        self.sha.update(result)
        self.left -= len(result)
        return result

    def rewind(self):
        self.fp.seek(self.position)
        self.left = self.size
        self.sha = hashlib.sha512()


def split_file_into_chunks(fp, file_size, chunk_size):
    """
    Split file-like object into sequence of file-like objects, each of
    those with size no greater than chunk_size bytes.
    Those are just wrappers to the original file-like objects. They should be fully
    read sequentially, and they cannot be used concurrently with
    the original object.
    """
    position = 0
    while position < file_size:
        yield _FileChunk(fp, position, min(chunk_size, file_size - position))
        position += chunk_size


def get_file_size(fp):
    """Get size of the file or length of a bytes object"""
    fp.seek(0, 2)  # move the current position to the end of the file
    size = fp.tell()
    fp.seek(0, 0)  # move the current position to the beginning of the file
    return size


def date_format(date):
    if isinstance(date, (datetime.datetime, datetime.date)):
        return date.strftime("%Y-%m-%d")
    else:
        return date


def encode_path(path):
    if isinstance(path, text_type):
        path = path.encode('utf-8')
    return quote(path, b'/')


class FileDownload(object):
    """
    Provides the file length and other metadata.
    Delegates reads to underlying request's response.
    """

    def __init__(self, response, file):
        self.response = response
        self.file = file
        self.closed = False

    def __len__(self):
        return int(self.response.headers['content-length'])

    def write_to(self, fp, progress_callback=None):
        """
        Copy data to a file, then close the source.
        Optional progress_callback should have the signature of ProgressCallbacks.download_progress
        """
        downloaded = 0
        with self:
            for chunk in self.iter_content():
                fp.write(chunk)
                if progress_callback is not None:
                    downloaded += len(chunk)
                    progress_callback(self.file, self.file.size, downloaded)

    def save_to(self, path, progress_callback=None):
        """
        Create a new file and save the contents
        Optional progress_callback should have the signature of ProgressCallbacks.download_progress
        """
        with open(path, "wb") as fp:
            self.write_to(fp, progress_callback)

    def close(self):
        if not self.closed:
            self.response.close()

    def read(self, size=None, decode_content=True):
        """
        Wrap urllib3 response.
        size - How much of the content to read. If specified, caching is skipped because it doesn't make sense to cache partial content as the full response.
        decode_content - If True, will attempt to decode the body based on the 'content-encoding' header.
        """
        return self.response.raw.read(size, decode_content)

    def __iter__(self, **kwargs):
        """
        Iterate response body line by line.
        You can specify alternate delimiter with delimiter parameter.
        """
        return self.response.iter_lines(**kwargs)

    def iter_content(self, chunk_size=16 * 1024):
        return self.response.iter_content(chunk_size)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

DEFAULT_EXCLUDES = fnmatch.translate(".*")
DEFAULT_EXCLUDES_RE = re.compile(DEFAULT_EXCLUDES).match


def make_excluded(excludes=None):
    if excludes is None:
        return DEFAULT_EXCLUDES_RE
    patterns = [DEFAULT_EXCLUDES]
    patterns.extend(fnmatch.translated(x) for x in excludes)
    return re.compile("|".join(patterns)).match


def generate_paths(roots, excludes=None):
    """
    Walk set of paths in local filesystem, and for each file and directory generate a tuple of
    (is directory, absolute path, path relative root used to get to that file)
    """
    excluded = make_excluded(excludes)
    for root in roots:
        base = os.path.basename(root)
        if not excluded(base):
            is_dir = os.path.isdir(root)
            yield is_dir, root, base
            if is_dir:
                prefix_len = len(os.path.dirname(root))
                for dirpath, dirnames, filenames in os.walk(root, topdown=True, followlinks=True):
                    relpath = dirpath[prefix_len:].strip('/')
                    for is_dir, names in ((False, filenames), (True, dirnames)):
                        for name in names:
                            if not excluded(name):
                                yield is_dir, os.path.join(dirpath, name), "%s/%s" % (relpath, name)


def filter_none_values(dict):
    """Return dictionary with values that are None filtered out"""
    return {k: v for (k, v) in dict.items() if v is not None}


class ResultList(list):
    """
    List with additional attributes representing a partial list of objects that exist in the cloud.
    total_count: Count of all objects that exist.
    offset: Starting index of this slice of results.
    """
    # TODO: make this more lazy?

    def __init__(self, data, total_count, offset):
        super(ResultList, self).__init__(data)
        self.total_count = total_count
        self.offset = offset



from egnyte.tests.config import EgnyteTestCase

TEST_FOLDER_NAME = 'permissions'
USER_NAME = 'test_permissions_user'
GROUP_NAME = 'test_permissions_group'


class TestDeprecatedPermissions(EgnyteTestCase):
    def setUp(self):
        super(TestDeprecatedPermissions, self).setUp()
        self.folder = self.root_folder.folder(TEST_FOLDER_NAME)
        self.user = self.egnyte.users.create(userName=USER_NAME,
                                             externalId=USER_NAME,
                                             email='test@example.com',
                                             familyName='Doe',
                                             givenName='John',
                                             active=False,
                                             sendInvite=False)
        self.group = self.egnyte.groups.create(GROUP_NAME, members=[self.user])

    def tearDown(self):
        self.user.delete()
        self.group.delete()
        super(TestDeprecatedPermissions, self).tearDown()

    def test_permissions_for_folder_creator(self):
        self.folder.create(True)
        permissions = self.folder.get_permissions()

        self.assertEquals(permissions.user_to_permission[self.config['login']], 'Owner')
        self.assertIn(self.config['login'], permissions.permission_to_owner['Owner']['users'])

    def test_effective_permissions(self):
        self.folder.create(True)

        effective = self.folder.get_effective_permissions(self.config['login'])
        self.assertEquals('Owner', effective)

    def test_set_user_permissions(self):
        self.folder.create(True)

        self.folder.set_permissions('Editor', users=[USER_NAME])
        permissions = self.folder.get_permissions()

        self.assertEquals(permissions.user_to_permission[USER_NAME], 'Editor')

    def test_set_group_permissions(self):
        self.folder.create(True)

        self.folder.set_permissions('Viewer', groups=[GROUP_NAME])
        permissions = self.folder.get_permissions()

        self.assertEquals(permissions.group_to_permission[GROUP_NAME], 'Viewer')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from egnyte import exc
from egnyte.tests.config import EgnyteTestCase

FOLDER_NAME = 'Iñtërnâtiônàlizætiøν☃ test'
DESTINATION_FOLDER_NAME = 'destination'
SUB_FOLDER_NAME = 'subfolder'
FILE_IN_FOLDER_NAME = 'test.txt'
FILE_CONTENT = b'TEST FILE CONTENT'


class TestFolders(EgnyteTestCase):
    def setUp(self):
        super(TestFolders, self).setUp()
        self.folder = self.root_folder.folder(FOLDER_NAME)
        self.destination = self.root_folder.folder(DESTINATION_FOLDER_NAME)
        self.file = self.folder.file(FILE_IN_FOLDER_NAME)

    def test_folder_create(self):
        self.folder.create()
        self.assertIsNone(self.folder.check())
        with self.assertRaises(exc.InsufficientPermissions):
            self.folder.create(False)

    def test_folder_recreate(self):
        self.folder.create()
        self.assertIsNone(self.folder.check())

        self.folder.create()
        self.assertIsNone(self.folder.check())

    def test_folder_cannot_recreate(self):
        self.folder.create()
        self.assertIsNone(self.folder.check())

        with self.assertRaises(exc.InsufficientPermissions):
            self.folder.create(False)

    def test_folder_delete(self):
        self.folder.create()
        self.assertIsNone(self.folder.check())

        self.folder.delete()
        with self.assertRaises(exc.NotFound):
            self.folder.check()

    def test_folder_move(self):
        self.folder.create()
        moved = self.folder.move(self.destination.path)
        self.assertEqual(moved.path, self.destination.path, "Moved folder path should be identical")
        with self.assertRaises(exc.NotFound):
            self.folder.check()

    def test_folder_copy(self):
        self.folder.create()
        copied = self.folder.copy(self.destination.path)

        self.assertEqual(copied.path, self.destination.path, "Copied folder path should be identical")
        self.assertIsNone(self.folder.check())
        self.assertIsNone(self.destination.check())

    def test_folder_list(self):
        self.folder.create()
        subfolder = self.folder.folder(SUB_FOLDER_NAME).create()
        _file = self.folder.file(FILE_IN_FOLDER_NAME)
        _file.upload(FILE_CONTENT)

        self.folder.list()

        folders_list = self.folder.folders
        self.assertEqual(1, len(folders_list), "There should be one subfolder")
        self.assertEqual(folders_list[0]._url, subfolder._url, "Subfolder URLs should be identical")

        files_list = self.folder.files
        self.assertEqual(1, len(files_list), "There should be one file")
        self.assertEqual(files_list[0]._url, _file._url, "File URLs should be identical")



from __future__ import unicode_literals

from egnyte import base, exc


class Audits(base.HasClient):
    """
    This resource is used to generated various kinds of audit reports.
    """
    _url_template = "pubapi/v1/audit/%(type)s"

    def _job_id(self, response):
        json = exc.accepted.check_json_response(response)
        return json['id']

    def logins(self, format, date_start, date_end, events, access_points=None, users=None):
        """
        Generate login report.
        Parameters:

        * format: 'csv' or 'json'
        * date_start: string in 'YYYY-MM-DD' format or datetime.date - first day report should cover
        * date_end: string in 'YYYY-MM-DD' format or datetime.date - last day report should cover
        * events: list of events to report on, at least one must be specified, allowed values: logins, logouts, account_lockouts, password_resets, failed_attempts

        Returns an AuditReport object.
        """
        json = dict(format=format,
                    date_start=base.date_format(date_start),
                    date_end=base.date_format(date_end),
                    events=list(events))
        if access_points:
            json['access_points'] = list(access_points)
        if users:
            json['users'] = list(users)
        url = self._client.get_url(self._url_template, type="logins")
        r = self._client.POST(url, json)
        return AuditReport(self._client, id=self._job_id(r), format=format, type='logins')

    def files(self, format, date_start, date_end, folders=None, file=None, users=None, transaction_type=None):
        """
        Generate files report.
        Parameters:

        * format: 'csv' or 'json'
        * date_start: string in 'YYYY-MM-DD' format or datetime.date - first day report should cover
        * date_end: string in 'YYYY-MM-DD' format or datetime.date - last day report should cover

        Returns an AuditReport object.
        """
        json = dict(format=format,
                    date_start=base.date_format(date_start),
                    date_end=base.date_format(date_end))
        if folders:
            json['folders'] = list(folders)
        if file:
            json['file'] = file
        if users:
            json['users'] = list(users)
        if transaction_type:
            json['transaction_type'] = list(transaction_type)
        url = self._client.get_url(self._url_template, type="files")
        r = self._client.POST(url, json)
        return AuditReport(self._client, id=self._job_id(r), format=format, type='files')

    def permissions(self, format, date_start, date_end, folders, assigners, assignee_users, assignee_groups):
        """
        Generate permissions report.
        Parameters:

        * format: 'csv' or 'json'
        * date_start: string in 'YYYY-MM-DD' format or datetime.date - first day report should cover
        * date_end: string in 'YYYY-MM-DD' format or datetime.date - last day report should cover

        Returns an AuditReport object.
        """
        json = dict(format=format,
                    date_start=base.date_format(date_start),
                    date_end=base.date_format(date_end),
                    folders=list(folders),
                    assigners=list(assigners),
                    assignee_users=list(assignee_users),
                    assignee_groups=list(assignee_groups))
        url = self._client.get_url(self._url_template, type="permissions")
        r = self._client.POST(url, json)
        return AuditReport(self._client, id=self._job_id(r), format=format, type='permissions')

    def get(self, id):
        """Get a previously generated report by its id"""
        return AuditReport(self._client, id=id)


class AuditReport(base.Resource):
    _url_template = "pubapi/v1/audit/jobs/%(id)s"
    _url_template_completed = "pubapi/v1/audit/%(type)s/%(id)s"
    status = 'running'

    def is_ready(self):
        """
        True if report is ready to be downloaded.
        Does a single API request.
        """
        r = self._client.GET(self._url)
        if r.status_code == 303:
            self.status == 'completed'
            return True
        exc.default.check_response(r)
        return False

    def wait(self, check_time=5.0):
        """
        Block until report is ready.
        Probably only useful for command line applications.
        """
        import time
        while not self.is_ready():
            time.sleep(check_time)

    def complete_url(self):
        url = self._client.get_url(self._url_template_completed, type=self.type, id=self.id)
        return url

    def download(self):
        r = self._client.GET(self.complete_url(), stream=True)
        exc.default.check_response(r)
        return base.FileDownload(r, None)

    def json(self):
        r = self._client.GET(self.complete_url())
        return exc.default.check_json_response(r)



import os.path
import json


def add_directory(filepath):
    """
    Add '~/.egnyte' in a platform independent way to a file path it it's relative.
    """
    if os.path.isabs(filepath):
        return filepath
    return os.path.join(os.path.expanduser('~'), '.egnyte', filepath)


def load(filename=None):
    """
    Load configuration from a JSON file.
    If filename is None, ~/.egnyte/config.json will be loaded.
    If filename is not an absolute path, it will be prefixed with ~/.egnyte/
    Returns loaded config as a dictionary on success and {} on failure.
    """
    filename = add_directory(filename or 'config.json')
    try:
        with open(filename, "r") as f:
            return json.load(f)
    except IOError:
        pass
    return {}


def save(config, filename=None):
    """
    Load configuration from a JSON file.
    If filename is not an absolute path, it will be prefixed with ~/.egnyte/
    """
    filename = add_directory(filename or 'config.json')
    directory = os.path.dirname(filename)
    if not os.path.exists(directory):
        os.makedirs(directory, 0o700)
    with open(filename, "w") as f:
        json.dump(config, f, indent=2, sort_keys=True)



from egnyte import exc
from egnyte.tests.config import EgnyteTestCase
from helpers import upload_file
import datetime

RECIPIENT_EMAIL_1 = 'test1@example.com'
RECIPIENT_EMAIL_2 = 'test2@example.com'
EGNYTE_FILE_NAME = '/sample.png'
FILE_NAME = 'FILE1.png'
ACCESSIBILITY_ANYONE = 'anyone'
ACCESSIBILITY_DOMAIN = 'domain'


class TestLinks(EgnyteTestCase):
    def setUp(self):
        super(TestLinks, self).setUp()
        self.filepath = self.root_folder.path + EGNYTE_FILE_NAME
        self.folder = self.root_folder.create()

    def test_create_file_link(self):
        uploaded_file = upload_file(self.egnyte, FILE_NAME, self.filepath)
        link = uploaded_file.link(ACCESSIBILITY_ANYONE)

        self.__verify_link(link[0], self.filepath, ACCESSIBILITY_ANYONE, 'file')

    def test_create_folder_link(self):
        link = self.folder.link(ACCESSIBILITY_DOMAIN)

        self.__verify_link(link[0], self.root_folder.path, ACCESSIBILITY_DOMAIN, 'folder')

    def test_delete_link(self):
        link = self.folder.link(ACCESSIBILITY_DOMAIN)

        self.__verify_link(link[0], self.root_folder.path, ACCESSIBILITY_DOMAIN, 'folder')

        link[0].delete()

        with self.assertRaises(exc.NotFound):
            link[0].check()

    def test_several_links(self):
        links = self.folder.link(ACCESSIBILITY_ANYONE,
                                 recipients=[RECIPIENT_EMAIL_1, RECIPIENT_EMAIL_2],
                                 send_email=False)

        self.assertEqual(links[0].path, links[1].path, "Both links should point to the same file")
        self.assertEqual((RECIPIENT_EMAIL_1,), tuple(links[0].recipients), "Link one should be for first email")
        self.assertNotEqual(links[0].id, links[1].id, "Links should have different ids")

        links[0].delete()

        self.assertIsNone(links[1].check())  # link two should still exist
        self.assertRaises(exc.NotFound, links[0].check)  # link one should no longer exist

    def test_list_links(self):
        links = self.egnyte.links
        all_links = links.list()
        tomorrow = datetime.datetime.now() + datetime.timedelta(1)
        future = links.list(created_after=tomorrow)

        self.assertEqual([], future, "List of links created after tomorrow should be empty")
        self.assertEqual(0, future.total_count, "Total count of links created after tomorrow should be 0")

        past = links.list(created_before=tomorrow)

        self.assertEqual(tuple(all_links), tuple(past),
                         "List of links created before tomorrow should include all links")

    def __verify_link(self, link, path, accessibility, link_type):
        self.assertEqual(link.path, path)
        self.assertEqual(link.accessibility, accessibility)
        self.assertEqual(link.type, link_type)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os.path

from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), "README.rst"), "rt") as f:
    long_description = f.read()

args = dict(
    name='egnyte',
    version='0.5.3',
    author='Egnyte',
    author_email='api-support@egnyte.com',
    license='MIT',
    description='Egnyte Public API SDK',
    long_description=long_description,
    zip_safe=True,
    packages=["egnyte", "egnyte.tests"],
    url='https://developers.egnyte.com/',
    include_package_data=True,
    install_requires=[
        "requests>=2.13.0",
        "six>=1.10.0"
    ],
    platforms=["Any"],
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    test_suite="egnyte",
    extras_require={
        'docs': ['sphinx', 'sphinx-argparse'],
    },
    # TODO: add an entrypoint for a console script, or do we stick with 'python -m egnyte'?
)

if __name__ == '__main__':
    setup(**args)



from egnyte.tests.config import EgnyteTestCase
from helpers import upload_file

EGNYTE_FILE_NAME = '/sample.png'
FILE_NAME = 'FILE1.png'
COMMENT = 'test comment'


class TestComments(EgnyteTestCase):
    def setUp(self):
        super(TestComments, self).setUp()
        self.filepath = self.root_folder.path + EGNYTE_FILE_NAME
        self.root_folder.create()

    def test_create_file_comment(self):
        uploaded_file = upload_file(self.egnyte, FILE_NAME, self.filepath)
        comment = uploaded_file.add_note(COMMENT)

        self.assertEqual(comment.file_path, self.filepath)
        self.assertEqual(comment.message, COMMENT)

    def test_delete_file_comment(self):
        uploaded_file = upload_file(self.egnyte, FILE_NAME, self.filepath)
        comment = uploaded_file.add_note(COMMENT)

        all_comments = self.egnyte.notes.list()

        self.assertIn(comment, all_comments)

        comment.delete()
        all_comments = self.egnyte.notes.list()

        self.assertNotIn(comment, all_comments)

    def test_list_comments(self):
        uploaded_file = upload_file(self.egnyte, FILE_NAME, self.filepath)
        comment = uploaded_file.add_note(COMMENT)

        all_comments = self.egnyte.notes.list()

        self.assertIn(comment, all_comments)
        self.assertEqual(uploaded_file, comment.get_file())

        comments = uploaded_file.get_notes()

        self.assertEqual(tuple(comments), (comment,))

        comment.delete()
        comments = uploaded_file.get_notes()

        self.assertEqual(tuple(comments), ())

        all_comments = self.egnyte.notes.list()

        self.assertNotIn(comment, all_comments)



import unittest
import csv
from os import unlink
from egnyte import configuration, client
from tempfile import NamedTemporaryFile

CONFIG_NAME = 'test_config.json'

EXPECTED_FILES_TITLE = ['This report reflects activity up to 2/24/17 23:59:59']
EXPECTED_FILES_HEADER = ['File/Folder', "Target Path/Link", 'User/Location',
                         'Transaction Type', 'Action Info', 'Access', 'Time', 'IP Address']
EXPECTED_LOGINS_TITLE = [
    'This report reflects activity up to 2/24/17 23:59:59']
EXPECTED_LOGINS_HEADER = [
    'User Name', 'Event', 'IP Address', 'Access', 'Time', 'Logout Time']


class TestAudits(unittest.TestCase):

    def setUp(self):
        self.config = configuration.load(CONFIG_NAME)
        self.egnyte = client.EgnyteClient(self.config)
        self.temp_file = NamedTemporaryFile(delete=False)

    def tearDown(self):
        unlink(self.temp_file.name)
        self.egnyte.close()
        del self.egnyte

    def test_create_and_download_files_audit_report_in_csv(self):
        audit_report = self.egnyte.audits.files(
            'csv', '2017-02-20', '2017-02-24')

        audit_report.wait()

        downloaded_file = audit_report.download()
        downloaded_file.save_to(self.temp_file.name)

        rows = self.__read_csv_file(self.temp_file.name)
        self.temp_file.close()

        self.assertEqual(EXPECTED_FILES_TITLE, rows[0])
        self.assertEqual(EXPECTED_FILES_HEADER, rows[1])

    def test_create_and_download_logins_report_in_csv(self):
        audit_report = self.egnyte.audits.logins(
            'csv', '2017-02-20', '2017-02-24', ['logins'])

        audit_report.wait()

        download_file = audit_report.download()
        download_file.save_to(self.temp_file.name)

        rows = self.__read_csv_file(self.temp_file.name)
        self.temp_file.close()

        self.assertEqual(EXPECTED_LOGINS_TITLE, rows[0])
        self.assertEqual(EXPECTED_LOGINS_HEADER, rows[1])

    def __read_csv_file(self, file_path):
        rows = []
        with open(file_path, 'rb') as csvfile:
            reader = csv.reader(csvfile)
            for row in reader:
                rows.append(row)
        return rows
